﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Text.RegularExpressions;

namespace WordTree
{
    /**
     * The main form of the Application.
     */
    public partial class Form1 : Form
    {
        private WordTree wordTree;
        private WordTree newWordTree;
        private WordTree wordTreeUncut;

        private DisplayList displayList;
        private DisplayList targetHelpList;

        private Register register = null;
        private static bool playAlphaAnimation = false;
        private static bool playMoveTreeAnimation = false;

		private LinkedList<PointF> searchPositions = null;
        private int searchListPos = 0;

        private static float param = 0.0f;
        private static float resetStep = 0.02f;
        private static float step = resetStep;

        private string[] currentSearchString;
        private string[] newSearchString;

        public Form1()
        {
            InitializeComponent();
        }

        private void rebuildWordTree()
        {
            if (register != null)
            {
                String[] searchTerms;
                string pattern = @"(\.)|(,)|(\s+)|(\n+)|(\r+)|(;)|(:)|(\()|(\))";
                Regex regex = new Regex(pattern);
                searchTerms = regex.Split(textBox.Text);


                List<string> searchTermsFixed = new List<string>(searchTerms.Length);

                foreach (string wort in searchTerms)
                {
                    if (wort.Length == 0 || wort[0] == ' ')
                        continue;
                    searchTermsFixed.Add(wort);
                }

                searchTerms = searchTermsFixed.ToArray();

                if (searchTerms == null || searchTerms.Length == 0)
                    return;

                LinkedList<String> results = register.search(searchTerms);

                if (results.Count > 0)
                {
                    WordTree.busy = true;

                    labelFound.Invoke(new MethodInvoker(delegate()
                    {
                        labelFound.Text = "";
                    }));
                    wordTreeUncut = wordTree = new WordTree(results.First.Value);
                    currentSearchString = searchTerms;
                    zoomFactor = 1;
                    zoomWindowLeft = 0;
                    zoomWindowTop = 0;
                    searchPositions = null;

                    IEnumerator<String> iter = results.GetEnumerator();

                    iter.MoveNext();

                    while (iter.MoveNext())
                        wordTree.addSentence(iter.Current);

                    wordTree.createFrequencySorting();
                    wordTree.calcRootPosition();
                    WordTree.bezierScaling = 1.0f / wordTree.numLeaves;

                    WordTree.busy = false;
                    panel1.Invalidate();
                }
                else
                {
                    labelFound.Invoke(new MethodInvoker(delegate()
                    {
                        labelFound.Text = "Nichts gefunden!";
                    }));    
                }
            }



            
        }

        private void button1_Click(object sender, EventArgs e)
        {
            if (playAlphaAnimation)
                return;

            wordTree = new WordTree(textBox.Text);
            panel1.Invalidate();
        }

        private void Form1_Paint(object sender, PaintEventArgs e)
        {

        }

        private void button2_Click(object sender, EventArgs e)
        {
            if (playAlphaAnimation)
                return;

            String[] words = textBox.Text.Split(new char[] { ' ' });
            wordTree.addSentence(words);
            panel1.Invalidate();
        }

        private void Form1_SizeChanged(object sender, EventArgs e)
        {
            panel1.Invalidate();
        }

        private void beendenToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void öffnenToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (playAlphaAnimation)
                return;

            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {
                StreamReader sr = new StreamReader(openFileDialog.FileName);
    
                string pattern = @"(\.)|(,)|(\s+)|(\n+)|(\r+)|(;)|(:)|(\()|(\))";
                Regex regex = new Regex(pattern);
                string[] woerter = regex.Split(sr.ReadToEnd());
                sr.Close();

                List<string> woerterFixed = new List<string>(woerter.Length);

                foreach (string wort in woerter)
                {
                    if (wort.Length == 0 || wort[0] == ' ')
                        continue;
                    woerterFixed.Add(wort);
                }

                woerter = woerterFixed.ToArray();

                register = new Register(woerter);

                Graphics panelGraphics = panel1.CreateGraphics();
                register.renderTopWords(panelGraphics, panel1.Width - 1, panel1.Height - 1);
                wordTree = null;
                panel1.Invalidate();
            }
        }

        private void button3_Click(object sender, EventArgs e)
        {
            if (playAlphaAnimation)
                return;

            rebuildWordTree();
        }
      
        float zoomFactor = 1;
        float zoomWindowLeft = 0;
        float zoomWindowTop = 0;

        private void panel1_Paint(object sender, PaintEventArgs e)
        {
            if (wordTree == null)
            {
                if(register!=null)
                    register.renderTopWords(e.Graphics, panel1.Width - 1, panel1.Height - 1);
                    
                return;
            }
            
            e.Graphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;

            //no animation, standard rendering:
            if (!playAlphaAnimation && !playMoveTreeAnimation)
            {
                wordTree.render(e.Graphics, panel1.Width - 1, panel1.Height - 1, zoomFactor, zoomWindowLeft, zoomWindowTop, null, true, false);

                try
                {
                 if(wordTree==wordTreeUncut)
                     e.Graphics.DrawString(wordTreeUncut.root.leafCount.ToString() + " Vorkommnisse gefunden, " + wordTree.wordsRendered.ToString() + " Wörter gerendert", new Font(WordTree.font.FontFamily, 10.0f), WordTree.brush, new Point(0, 0));
                 else
                     e.Graphics.DrawString(wordTree.root.leafCount.ToString() + " von " + wordTreeUncut.root.leafCount.ToString() + " Vorkommnissen angezeigt, " + wordTree.wordsRendered.ToString() + " Wörter gerendert", new Font(WordTree.font.FontFamily, 10.0f), WordTree.brush, new Point(0, 0));          

                }
                catch (Exception)
                { }
                
                return;
            }

            if (wordTree.lastSplitNode.leafCount < wordTree.numLeaves)
            {
                //alpha animation going on:
                if (playAlphaAnimation)
                {
                    float diminishingFactor;
                    if (wordTree.numLeaves < 100)
                        diminishingFactor = 0.8f;
                    else if (wordTree.numLeaves < 300)
                        diminishingFactor = 0.6f;
                    else if (wordTree.numLeaves < 1000)
                        diminishingFactor = 0.5f;
                    else if (wordTree.numLeaves < 5000)
                        diminishingFactor = 0.3f;
                    else
                        diminishingFactor = 0.1f;

                    wordTree.animationAlpha = (int)((float)wordTree.animationAlpha * diminishingFactor);
                    wordTree.render(e.Graphics, panel1.Width - 1, panel1.Height - 1, zoomFactor, zoomWindowLeft, zoomWindowTop, null, true, false);
                    panel1.Invalidate();

                    if (wordTree.animationAlpha <= 1)
                    {
                        playAlphaAnimation = false;
                        wordTree.render(e.Graphics, panel1.Width - 1, panel1.Height - 1, zoomFactor, zoomWindowLeft, zoomWindowTop, displayList, false, false);

                        targetHelpList = new DisplayList();
                        newWordTree.render(e.Graphics, panel1.Width - 1, panel1.Height - 1, zoomFactor, zoomWindowLeft, zoomWindowTop, targetHelpList, false, false);

                        displayList.mergeTargetList(targetHelpList);

                        playMoveTreeAnimation = true;
                        param = 0.0f;
                        panel1.Invalidate();

                        return;

                    }
                }

                if (playMoveTreeAnimation)
                {

                    step = Math.Max(0.0005f, (1.0f - param) * resetStep);

                    param += step;

                    if (param < 1.0f)
                        displayList.update(param);
                    else
                    {
                        playMoveTreeAnimation = false;
                        step = resetStep;
                        wordTree = newWordTree;
                        wordTreeUncut = wordTree;
                        currentSearchString = newSearchString;
                    }

                    displayList.render(e.Graphics, panel1.Width - 1, panel1.Height - 1, zoomFactor, zoomWindowLeft, zoomWindowTop);
                    panel1.Invalidate();

                    return;
                }
            }
            else
            {
                if (playMoveTreeAnimation)
                {
                    step = Math.Max(0.0005f, (1.0f - param) * resetStep);

                    param += step;

                    if (param < 1.0f)
                        displayList.update(param);
                    else
                    {
                        playMoveTreeAnimation = false;
                        step = resetStep;
                        newWordTree.animationAlpha = 10;
                        playAlphaAnimation = true;
                    }

                    displayList.render(e.Graphics, panel1.Width - 1, panel1.Height - 1, zoomFactor, zoomWindowLeft, zoomWindowTop);
                    panel1.Invalidate();

                    return;
                }

                if (playAlphaAnimation)
                {
                    float invDiminishingFactor;
                    if (wordTree.numLeaves < 100)
                        invDiminishingFactor = 0.8f;
                    else if (wordTree.numLeaves < 300)
                        invDiminishingFactor = 0.6f;
                    else if (wordTree.numLeaves < 1000)
                        invDiminishingFactor = 0.5f;
                    else if (wordTree.numLeaves < 5000)
                        invDiminishingFactor = 0.3f;
                    else
                        invDiminishingFactor = 0.1f;

                    invDiminishingFactor = 1.0f / invDiminishingFactor;

                    newWordTree.animationAlpha = (int)Math.Ceiling((float)newWordTree.animationAlpha * invDiminishingFactor);
                    if (newWordTree.animationAlpha > 250)
                        newWordTree.animationAlpha = 255;
                    newWordTree.render(e.Graphics, panel1.Width - 1, panel1.Height - 1, zoomFactor, zoomWindowLeft, zoomWindowTop, null, true, true);
                    panel1.Invalidate();
                    if (newWordTree.animationAlpha > 250)
                    {
                        step = resetStep;
                        wordTree = newWordTree;
                        wordTreeUncut = wordTree;
                        wordTree.animationAlpha = 255;
                        playAlphaAnimation = false;
                        currentSearchString = newSearchString;
                    }
                }
            }
             
        }

        int oldMousePosX, oldMousePosY;
        private void panel1_MouseMove(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
            {
                zoomWindowLeft -= (e.X - oldMousePosX) / zoomFactor;
                zoomWindowTop -= (e.Y - oldMousePosY) / zoomFactor;
                panel1.Invalidate();
            }
            oldMousePosX = e.X;
            oldMousePosY = e.Y;
        }

        private void panel1_MouseWheel(object sender, MouseEventArgs e)
        {
            zoomFactor *= 1.0f + e.Delta / 2000.0f;
            if (zoomFactor < 0.000001f)
                zoomFactor = 0.000001f;
            panel1.Invalidate();
        }

        private void Form1_Resize(object sender, EventArgs e)
        {
            button3.Left = (this.Width-button3.Width) / 2;


            panel1.Invalidate();
        }

        private void leafCountAnzeigenToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (playAlphaAnimation)
                return;

            WordTree.showLeafCount = !WordTree.showLeafCount;
            panel1.Invalidate();
        }

        private void panel1_MouseClick(object sender, MouseEventArgs e)
        {
            if (playAlphaAnimation || playMoveTreeAnimation || wordTree == null || register == null)
                return;
            
            bool ctrl_pressed = false;

            if (Control.ModifierKeys == Keys.Control)
                ctrl_pressed = true;

            if (e.Button == MouseButtons.Left)
            {
                Graphics panelGraphics = panel1.CreateGraphics();
                
                LinkedList<string> newSearchString = wordTree.getSubtreeStringList(panelGraphics, e.X, e.Y, panel1.Width, panel1.Height, zoomFactor, zoomWindowLeft, zoomWindowTop);

                if (newSearchString == null)
                {
                    wordTree.lastSplitNode = null; //invalidate last split node, since no splitting will take place
                    return;
                }

            

                if (ctrl_pressed)
                {
                    string last = newSearchString.Last();
                    newSearchString.Clear();
                    newSearchString.AddFirst(last);
                }

                this.newSearchString = newSearchString.ToArray<string>();

                LinkedList<String> results = register.search(this.newSearchString);

                if (results.Count == 0)
                    MessageBox.Show("Keine Vorkommnisse im Text gefunden.");
                else
                {
                    //create newWordTree
                    newWordTree = new WordTree(results.First.Value);
                    IEnumerator<String> iter = results.GetEnumerator();
                    iter.MoveNext();

                    while (iter.MoveNext())
                        newWordTree.addSentence(iter.Current);

                    newWordTree.createFrequencySorting();
                    newWordTree.rootPosition = new Point(0, (int)wordTree.lastSplitNode.getLastPosition().Y);

                    if (!ctrl_pressed)
                    {
                        displayList = new DisplayList();
                        wordTree.resetSideNodeMarking();
                        if (wordTree.lastSplitNode.leafCount < wordTree.numLeaves)
                        {
                            playAlphaAnimation = true;
                        }
                        else
                        {
                            playMoveTreeAnimation = true;
                            newWordTree.markAllSideNodes(currentSearchString);
                            newWordTree.render(panel1.CreateGraphics(), panel1.Width - 1, panel1.Height - 1, zoomFactor, zoomWindowLeft, zoomWindowTop, displayList, false, false);

                            targetHelpList = new DisplayList();
                            wordTree.render(panel1.CreateGraphics(), panel1.Width - 1, panel1.Height - 1, zoomFactor, zoomWindowLeft, zoomWindowTop, targetHelpList, false, false);

                            displayList.mergeTargetList(targetHelpList);
                            displayList.reverseSourceTarget();

                            playMoveTreeAnimation = true;
                            param = 0.0f;
                        }
                        

                    }
                    else
                    {
                        wordTree = newWordTree;
                        wordTreeUncut = wordTree;
                    }

                    panel1.Invalidate();
                }
            }
        }

      private void textBox_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (playAlphaAnimation || playMoveTreeAnimation)
                return;

            if (e.KeyChar == 32 || e.KeyChar == 13)
            {
                while (WordTree.busy)
                    ;//wait
                    
                System.Threading.Thread t = new System.Threading.Thread(rebuildWordTree);
                t.Priority = System.Threading.ThreadPriority.Lowest;
                t.Start();
            }
        }

      private void vorkommnisToolStripMenuItem_Click(object sender, EventArgs e)
      {
          if (playAlphaAnimation || playMoveTreeAnimation)
              return;

          wordTree.resort(0);
          panel1.Invalidate();
      }

      private void alphabetischToolStripMenuItem_Click(object sender, EventArgs e)
      {
          if (playAlphaAnimation || playMoveTreeAnimation)
              return;

          wordTree.resort(1);
          panel1.Invalidate();
      }

      private void häufigkeitToolStripMenuItem_Click(object sender, EventArgs e)
      {
          if (playAlphaAnimation || playMoveTreeAnimation)
              return;

          wordTree.resort(2);
          panel1.Invalidate();
      }

      private void schriftartToolStripMenuItem_Click(object sender, EventArgs e)
      {
          if (playAlphaAnimation || playMoveTreeAnimation)
              return;

          if  (fontDialog1.ShowDialog() == DialogResult.OK)
          {
              try
              {
                  Font f = new Font(fontDialog1.Font.FontFamily, 10.0f);
              }
              catch (Exception)
              {
                  return;
              }

             WordTree.font = fontDialog1.Font;
             panel1.Invalidate();
          }
      }
    private void suchenToolStripMenuItem_Click(object sender, EventArgs e)
      {
          Suchfenster sf = new Suchfenster();

          sf.ShowDialog();

          string[] terms = sf.searchStringTextBox.Text.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);

          if (terms.Length > 0 && wordTree != null)
          {
              // Stelle sicher, dass alle die richtige Position haben (nicht geclipped)
              wordTree.render(panel1.CreateGraphics(), panel1.Width - 1, panel1.Height - 1, zoomFactor, zoomWindowLeft, zoomWindowTop, new DisplayList(), false, false);
              searchPositions = wordTree.searchWord(terms[0]);
              searchListPos = -1;
              gotoNextSearchPosition();
              panel1.Invalidate();
          }
      }

      private void gotoNextSearchPosition()
      {
          if (searchPositions != null && searchListPos < searchPositions.Count - 1)
          {
              searchListPos++;
              zoomWindowLeft = searchPositions.ElementAt(searchListPos).X - ((panel1.Width / 2));
              zoomWindowTop = searchPositions.ElementAt(searchListPos).Y - ((panel1.Height / 2));
          }
      }

      private void gotoPreviousSearchPosition()
      {
          if (searchPositions != null && searchListPos > 0)
          {
              searchListPos--;
              zoomWindowLeft = searchPositions.ElementAt(searchListPos).X - ((panel1.Width / 2));
              zoomWindowTop = searchPositions.ElementAt(searchListPos).Y - ((panel1.Height / 2));
          }
      }

      private void nToolStripMenuItem_Click(object sender, EventArgs e)
      {
          gotoNextSearchPosition();
          panel1.Invalidate();
      }

      private void vorherigesResultatToolStripMenuItem_Click(object sender, EventArgs e)
      {
          gotoPreviousSearchPosition();
          panel1.Invalidate();
      }

      private void bearbeitenToolStripMenuItem_Click(object sender, EventArgs e)
      {

      }

      private void bezierkurvenToolStripMenuItem_Click(object sender, EventArgs e)
      {
          new Form2(panel1, WordTree.bezierScaling).ShowDialog();
      }

      private void ganzenBaumZeigenToolStripMenuItem_Click_1(object sender, EventArgs e)
      {
          if (wordTree == null || WordTree.font == null)
              return;

          float yMax = (int)(wordTree.root.leafCount * WordTree.font.Size * 2 * WordTree.fontScale);

          zoomFactor = panel1.Height / (1.1f * yMax);

          zoomWindowLeft = (wordTree.rootPosition.X + (0.5f * panel1.Width) / zoomFactor) - 0.5f * panel1.Width;
          zoomWindowTop = (wordTree.rootPosition.Y + 0.5f * WordTree.font.Size) - (0.5f * panel1.Height);

          panel1.Invalidate();
      }

      private void zurWurzelSpringenToolStripMenuItem_Click_1(object sender, EventArgs e)
      {
          if (wordTree == null)
              return;

          zoomFactor = 1;
          zoomWindowLeft = wordTree.rootPosition.X;
          zoomWindowTop = wordTree.rootPosition.Y;
          panel1.Invalidate();
      }

      private void prozentDerVorkommnisseToolStripMenuItem5_Click(object sender, EventArgs e)
      {
          wordTree = wordTreeUncut;
          panel1.Invalidate();
      }


      private void prozentDerVorkommnisseToolStripMenuItem_Click(object sender, EventArgs e)
      {
          wordTree = wordTreeUncut.removeUnimportantNodes(1);
          wordTree.createFrequencySorting();
          panel1.Invalidate();
      }

      private void prozentDerVorkommnisseToolStripMenuItem1_Click(object sender, EventArgs e)
      {
          wordTree = wordTreeUncut.removeUnimportantNodes(2);
          wordTree.createFrequencySorting();
          panel1.Invalidate();
      }

      private void prozentDerVorkommnisseToolStripMenuItem2_Click(object sender, EventArgs e)
      {
          wordTree = wordTreeUncut.removeUnimportantNodes(3);
          wordTree.createFrequencySorting();
          panel1.Invalidate();
      }

      private void prozentDerVorkommnisseToolStripMenuItem3_Click(object sender, EventArgs e)
      {
          wordTree = wordTreeUncut.removeUnimportantNodes(4);
          wordTree.createFrequencySorting();
          panel1.Invalidate();
      }

      private void prozentDerVorkommnisseToolStripMenuItem4_Click(object sender, EventArgs e)
      {
          wordTree = wordTreeUncut.removeUnimportantNodes(5);
          wordTree.createFrequencySorting();
          panel1.Invalidate();
      }

      private void relativZurEbeneToolStripMenuItem_Click(object sender, EventArgs e)
      {
          WordTree.fontSizeCoding = 1;
          panel1.Invalidate();
      }

      private void relativZurGesamtheitToolStripMenuItem_Click(object sender, EventArgs e)
      {
          WordTree.fontSizeCoding = 2;
          panel1.Invalidate();
      }
    
    }
}
        