﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;

namespace WordTree
{
    /**
     * A quick data structure for look-ups.
     */
    class Register
    {
        private Dictionary<string, LinkedList<long>> wortPosList;
        private List<KeyValuePair<string, int>> frequencyCountList;
        private List<KeyValuePair<string, int>> topWords;
        private String[] woerter;
        private long anzWords;
        private static int numTopWords = 30;

        private static string[] stopwords = { "\n", "\r", ",", ".", ":", "(", ")", ";", "a", "about", "above", "above", "across", "after", "afterwards", "again", "against", "all", "almost", "alone", "along", "already", "also", "although", "always", "am", "among", "amongst", "amoungst", "amount", "an", "and", "another", "any", "anyhow", "anyone", "anything", "anyway", "anywhere", "are", "around", "as", "at", "back", "be", "became", "because", "become", "becomes", "becoming", "been", "before", "beforehand", "behind", "being", "below", "beside", "besides", "between", "beyond", "bill", "both", "bottom", "but", "by", "call", "can", "cannot", "cant", "co", "con", "could", "couldnt", "cry", "de", "describe", "detail", "do", "done", "down", "due", "during", "each", "eg", "eight", "either", "eleven", "else", "elsewhere", "empty", "enough", "etc", "even", "ever", "every", "everyone", "everything", "everywhere", "except", "few", "fifteen", "fify", "fill", "find", "fire", "first", "five", "for", "former", "formerly", "forty", "found", "four", "from", "front", "full", "further", "get", "give", "go", "had", "has", "hasnt", "have", "he", "hence", "her", "here", "hereafter", "hereby", "herein", "hereupon", "hers", "herself", "him", "himself", "his", "how", "however", "hundred", "i", "ie", "if", "in", "inc", "indeed", "interest", "into", "is", "it", "its", "itself", "keep", "last", "latter", "latterly", "least", "less", "ltd", "made", "many", "may", "me", "meanwhile", "might", "mill", "mine", "more", "moreover", "most", "mostly", "move", "much", "must", "my", "myself", "name", "namely", "neither", "never", "nevertheless", "next", "nine", "no", "nobody", "none", "noone", "nor", "not", "nothing", "now", "nowhere", "of", "off", "often", "on", "once", "one", "only", "onto", "or", "other", "others", "otherwise", "our", "ours", "ourselves", "out", "over", "own", "part", "per", "perhaps", "please", "put", "rather", "re", "same", "see", "seem", "seemed", "seeming", "seems", "serious", "several", "she", "should", "show", "side", "since", "sincere", "six", "sixty", "so", "some", "somehow", "someone", "something", "sometime", "sometimes", "somewhere", "still", "such", "system", "take", "ten", "than", "that", "the", "their", "them", "themselves", "then", "thence", "there", "thereafter", "thereby", "therefore", "therein", "thereupon", "these", "they", "thickv", "thin", "third", "this", "those", "though", "three", "through", "throughout", "thru", "thus", "to", "together", "too", "top", "toward", "towards", "twelve", "twenty", "two", "un", "under", "until", "up", "upon", "us", "very", "via", "was", "we", "well", "were", "what", "whatever", "when", "whence", "whenever", "where", "whereafter", "whereas", "whereby", "wherein", "whereupon", "wherever", "whether", "which", "while", "whither", "who", "whoever", "whole", "whom", "whose", "why", "will", "with", "within", "without", "would", "yet", "you", "your", "yours", "yourself", "yourselves", "the" };

        /**
         * Creates a register with the words in _woerter.
         */
        public Register(String[] _woerter)
        {
            woerter = _woerter;
            anzWords = woerter.Length;
            wortPosList = new Dictionary<string, LinkedList<long>>(new CaseInSensitiveEqualityComparer());

            for (long wort = 0; wort < woerter.Length; wort++)
            {
                LinkedList<long> posList = null;
                bool found = true;
                if (!wortPosList.TryGetValue(woerter[wort], out posList))
                {
                    posList = new LinkedList<long>();
                    found = false;
                }

                posList.AddLast(wort);
                if (!found)
                    wortPosList.Add(woerter[wort], posList);
            }

            frequencyCountList = new List<KeyValuePair<string, int>>(wortPosList.Count);
           
            foreach(KeyValuePair<string, LinkedList<long>> elem in wortPosList)
            {
                if(!isStopWord(elem.Key))
                    frequencyCountList.Add(new KeyValuePair<string, int>(elem.Key, elem.Value.Count));
                
            }

            frequencyCountList.Sort(delegate(KeyValuePair<string, int> firstPair, KeyValuePair<string, int> nextPair) { return -firstPair.Value.CompareTo(nextPair.Value); });

            topWords = new List<KeyValuePair<string, int>>(10);

            int i = 0;

            while(topWords.Count < numTopWords && i<frequencyCountList.Count)
            {
                KeyValuePair<string, int> testWord = frequencyCountList.ElementAt(i);

                if (!isStopWord(testWord.Key))
                    topWords.Add(testWord);

                i++;
            }            
        }

        /**
         * Searches the register for searchTerms (order dependant).
         */
        public LinkedList<String> search(String[] searchTerms)
        {
            if (searchTerms == null || searchTerms.Length == 0)
                return null;

            LinkedList<String> result = new LinkedList<String>();
            int checkUntil = searchTerms.Length;
            
            LinkedList<long> posList = null;
            if (wortPosList.TryGetValue(searchTerms[0], out posList))
            {
                foreach (long pos in posList)
                {
                    String sentence = searchTerms[0];

                    int c = 1;
                    bool mismatch = false;

                    for (long i = pos+1; i < pos+10; i++, c++)
                    {
                        if(c< checkUntil && !CaseInSensitiveEqualityComparer.EqualsStatic(woerter[i], searchTerms[c]))
                        {
                            mismatch = true;
                            break;
                        }

                        sentence += " " + woerter[i];
                        if (woerter[i].Contains('.') || woerter[i].Contains('!') || woerter[i].Contains('?'))
                            break;
                    }

                    if (mismatch)
                        continue;

                    result.AddLast(sentence);
                }
            }

            return result;
        }

        /**
         * Tells if the word in s is a stop word (i.e. a word which should not be counted).
         */
        private bool isStopWord(string s)
        {
            if(stopwords.Contains(s.ToLower()))
                return true;
            else
                return false;
        
        }
     
        public void renderTopWords(Graphics g, int panelWidth, int panelHeight)
        {
            int ySpace = panelHeight / numTopWords;
            int y = 0;

            int sum = 0;
            foreach (KeyValuePair<string, int> pair in topWords)
                sum += pair.Value;


            foreach(KeyValuePair<string, int> pair in topWords)
            {
                Font f = new Font(WordTree.font.FontFamily, panelHeight*((float)pair.Value / sum));

                if (f.Size > 1)
                    g.DrawString(pair.Key, f, WordTree.brush, new Point(10, y));
                y += (int)(f.Size);
            }
        }
    }


}
